/*
 * Copyright (c) 2004, 2005, 2006
 *      Luiz Souza <loos.br@gmail.com>.  All rights reserved.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL AUTHOR OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sys/types.h>
#include <sys/stat.h>

#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>
#include <pwd.h>

#include "string.h"
#include "syslog.h"
#include "vchkuser.h"

int make_path(string *alias, string *user, string *domain_path)
{
    if (str_copy(alias, domain_path->s, domain_path->len) == 0 ||
	str_cat(alias, "/.qmail-", 8) == 0 ||
	str_cat(alias, user->s, user->len) == 0)
	return(FAIL);
    return(OK);
}

int make_default_path(string *alias, string *user, string *domain_path)
{
    if (str_copy(alias, domain_path->s, domain_path->len) == 0 ||
	str_cat(alias, "/.qmail-", 8) == 0 ||
	str_cat(alias, user->s, user->len) == 0 ||
	str_cat(alias, "-default", 8) == 0)
	return(FAIL);
    return(OK);
}

int check_path(string *alias)
{
 struct stat	st;

    /* get stats from alias file */
    if (stat(alias->s, &st) == -1) {
	switch (errno) {
	    case ENOENT:
		/* .qmail doesnt exist? refuse address */
		return(BOUNCE);

	    case EACCES:
		/* can't access .qmail - let vpopmail deal with it */
		vlog("address bypass - wrong perms for %s", alias->s);
		return(NOTBOUNCE);

	    default:
		vlog("fail - stat");
		return(FAIL);
	}
    }

    /* check if alias file is a regular file or a symlink */
    if (!S_ISREG(st.st_mode) && !S_ISLNK(st.st_mode))
	return(BOUNCE);

    /*
     * check access rights to alias
     */
    if (!(st.st_mode & S_IRUSR)) {
	vlog("address bypass - wrong perms for %s", alias->s);
	return(NOTBOUNCE);
    }

    vlog("address accepted - alias found");
    return(NOTBOUNCE);
}

int valias_check(string *user, string *domain_path)
{
 string		alias_path;
 int		err = BOUNCE;			/* default action */

    /* zero alias_path */
    str_zero(&alias_path);

    /* look for .qmail-user */
    if (make_path(&alias_path, user, domain_path) == FAIL)
	goto fail_nomem;

    /* check if exist */
    err = check_path(&alias_path);
    if (err != BOUNCE) {
	str_free(&alias_path);
	return(err);
    }

    /* look for .qmail-user-default */
    if (make_default_path(&alias_path, user, domain_path) == -1)
	goto fail_nomem;

    err = check_path(&alias_path);
    str_free(&alias_path);
    return(err);

fail_nomem:
    vlog("fail - no memory");
    str_free(&alias_path);
    return(FAIL);
}

int valias_make(string *alias, string *user)
{

    /* alloc alias */
    if (str_ready(alias, user->len + 1) == 0)
	return(FAIL);

    /* copy and replace any dot on user */
    user->p = user->s;
    while (*(user->p) && alias->len < alias->a) {
	if (*(user->p) == '.')
	    alias->s[alias->len++] = ':';
	else
	    alias->s[alias->len++] = *(user->p);
	user->p++;
    }

    /* safety */
    alias->s[alias->len] = 0;

    return(OK);
}

int valias(string *user, string *domain_path)
{
 int		err = BOUNCE;
 string		alias;

    /* zero alias */
    str_zero(&alias);

    /* replace any '.' on user by ':' */
    if (valias_make(&alias, user) == FAIL)
	return(FAIL);

    /* check for user */
    err = valias_check(&alias, domain_path);
    if (err != BOUNCE) {
	str_free(&alias);
	return(err);
    }

    /* check for user-ext */
    alias.p = (unsigned char *)strrchr(alias.s, '-');
    if (alias.p == NULL) {
	/* no user-ext */
	str_free(&alias);
	return(BOUNCE);
    }

    /* set len for user */
    if (alias.len < (alias.p - alias.s)) {
	/* safety */
	str_free(&alias);
	return(BOUNCE);
    }
    alias.len = alias.p - alias.s;

    err = valias_check(&alias, domain_path);
    str_free(&alias);
    return(err);
}
